/* written by John Doug Reynolds, May 2002 */

#include "rgsspectrum.h"
#include <ShapeScanner.h>
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <CalServer.h>
#include <ASCregion.h>
#include <cstdio>

namespace {
   using namespace rgsPixelsLib;

   class Statistics {
   public:
      float _mass, _exposure;
      unsigned _counts, _pixels;
      Statistics();
      Statistics& operator +=( const Statistics& );
   };

   class AccumSpan : public Span {
   public:
      AccumSpan( const Span& );
      void expose( const ExposureMap&, int column, int offset );
      void expose( float exposure ); // emergency kludge
      bool accept( short xdsp );
      struct Statistics stats( int rebin ) const;
   private:
      vector<unsigned> _counts;
      vector<float> _exposure;
   };

   class Accumulator {
   public:
      void spans( const list<Span>& );
      void expose( const ExposureMap&, int column, int offset );
      void expose( float exposure ); // emergency kludge
      bool accept( short xdsp );
      struct Statistics stats( int rebin ) const;
   private:
      list<AccumSpan> _spans;
   };

   class SpectralBin {
   public:
      struct Accumulator foreground, background;
      int nodeId; // emergency kludge
      bool dubious;

      SpectralBin() : nodeId(-1), dubious(false) {}
      void phSpans( const list<Span> &s ) { _banana = s; }
      void fgSpans( const list<Span> &s ) { foreground.spans(s); }
      void bgSpans( const list<Span> &s ) { background.spans(s); }
      bool accept( short xdsp, short calpi );

   private:
      list<Span> _banana;
   };

   class Machine {
   public:
      Machine( const GridSpan &beta, const GridSpan &xdsp,
	       int source, int order, int rebin );

      void loadRegions( struct DataSet *sources, bool withfg, bool withbg );
      void loadExposure( struct Array *emap, int edgechannels );
      void loadEvents( const Table *events );

      void out( const string &dsname, const string &hduclas2,
		const list<const Attribute*>& ) const;
   private:
      typedef void (SpectralBin::*spaninit_f)( const list<Span>& );

      struct GridSpan _grid, _betaGrid, _xdspGrid;
      double _maxExposure, _lineSep, _rgsAlpha;
      int _source, _order, _rebin;
      vector<SpectralBin> _accum;

      void scan( struct Table*, spaninit_f, bool spatial );
      void out( struct DataSet*, const string &hduclas2,
		const list<const Attribute*>& ) const;
   };
}

void rgsspectrum()
{
   bool bgcorrect = false;
   string fgsetname, bgsetname;

   string dsname	= stringParameter("evlist");
   string sdefsetname	= stringParameter("srclist");
   bool withfgset	= booleanParameter("withspectrum");
   if ( withfgset ) {
      bgcorrect		= booleanParameter("bkgcorrect");
      fgsetname		= stringParameter("spectrumset");
   }
   bool withbgset	= booleanParameter("withbkgset");
   if ( withbgset )
      bgsetname		= stringParameter("bkgset");
   int source		= intParameter("source");
   int order		= intParameter("order");
   int rebin		= intParameter("rebin");
   int edgechannels	= intParameter("edgechannels");

   struct DataSet *dataset = dataSetServer->open( dsname, Dal::Read );

   enum Xmm::DataMode datamode
      = Xmm::toDataMode( dataset->attribute("DATAMODE")->asString() );
   if ( datamode == Xmm::HIGH_TIME_RESOLUTION ) {
      if ( bgcorrect ) {
	 errstr << "parameter --bkgcorrect=true not implemented for HTR"
		<< Sas::warning("DisabledParameter");
	 bgcorrect = false;
      }
      withbgset = false;
   }

   if ( !withfgset && !withbgset ) {
      Sas::warning("NothingRequested","both outputs have been disabled");
      dataSetServer->close( dataset );
      return;
   }

   const Table *etab = dataset->table("EVENTS");
   if ( source == 0 ) source = etab->attribute("SOURCEID")->asInt();

   char scratch[32];
   string::size_type suffix = dsname.rfind('.');
   if ( suffix == string::npos ) suffix = dsname.size();

   if ( sdefsetname == "" )
      if ( suffix >= 10 )
	 (sdefsetname=dsname).replace( suffix-10, 6, "SRCLI_" );
      else
	 sdefsetname = string("srclist") + dsname.substr(suffix);

   if ( withfgset && fgsetname == "" )
      if ( suffix >= 10 ) {
	 sprintf( scratch, "SRSPEC%d%03d", order, source );
	 (fgsetname=dsname).replace( suffix-10, 10, scratch );
      }
      else {
	 sprintf( scratch, "src%do%d.pha", source, order );
	 fgsetname = scratch;
      }

   if ( withbgset && bgsetname == "" )
      if ( suffix >= 10 ) {
	 sprintf( scratch, "BGSPEC%d%03d", order, source );
	 (bgsetname=dsname).replace( suffix-10, 10, scratch );
      }
      else {
	 sprintf( scratch, "bkg%do%d.pha", source, order );
	 bgsetname = scratch;
      }

   struct Machine m( gridspan( etab->column("BETA_CHANNEL") ),
		     gridspan( etab->column("XDSP_CHANNEL") ),
		     source, order, rebin
		     );

   struct DataSet *sdefset = dataSetServer->open( sdefsetname, Dal::Read );
   m.loadRegions( sdefset, withfgset, (withfgset&&bgcorrect)||withbgset );
   dataSetServer->close( sdefset );

   for ( unsigned ccd = 1; ccd <= 9; ++ccd )
      for ( unsigned node = 0; node < 2; ++node ) {
	 char extname[9];
	 sprintf( extname, "EXPMAP%u%u", node, ccd );
	 if ( dataset->hasBlock(extname) )
	    m.loadExposure( dataset->array(extname), edgechannels );
      }

   m.loadEvents( etab );

   list<const Attribute*> attrib;
   attrib.push_back( dataset->attribute("TELESCOP") );
   attrib.push_back( dataset->attribute("INSTRUME") );
   attrib.push_back( dataset->attribute("FILTER") );
   attrib.push_back( dataset->attribute("OBJECT") );
   attrib.push_back( dataset->attribute("OBS_ID") );
   attrib.push_back( dataset->attribute("EXP_ID") );
   attrib.push_back( dataset->attribute("DATE-OBS") );
   attrib.push_back( dataset->attribute("DATE-END") );
   attrib.push_back( dataset->attribute("EQUINOX") );
   attrib.push_back( dataset->attribute("RADECSYS") );
   attrib.push_back( etab->attribute("RA_OBJ") );
   attrib.push_back( etab->attribute("DEC_OBJ") );

   if ( withfgset )
      m.out( fgsetname, bgcorrect ? "NET" : "TOTAL", attrib );

   if ( withbgset )
      m.out( bgsetname, "BKG", attrib );

   dataSetServer->close( dataset );
}

/* --- Machine methods ----------------------------------------------------- */

Machine::Machine( const GridSpan &beta, const GridSpan &xdsp,
		  int source, int order, int rebin )
   :
   _betaGrid(beta), _xdspGrid(xdsp), _maxExposure(0),
   _source(source), _order(order), _rebin(rebin)
{
   // rebin: minval remains same, maxval grows as necessary
   size_t nbins = (size_t) ceil( beta.length() / (float)rebin );
   _grid = GridSpan( Grid( rebin*beta.stride, beta[(.5-beta.min)*(rebin-1)] ),
		     Span(beta.min).length(nbins) );
   _accum.resize(nbins);
}

void Machine::scan( struct Table *table, spaninit_f initf, bool spatial )
{
   struct InASCregionTable ascReg(table);
   struct GeoRegion geoReg;
   ascReg >> geoReg;

   struct RegionScanner scanner( geoReg, _grid, spatial ? _xdspGrid : Grid() );
   struct Bounds clip(_grid,spatial ? (Span&)_xdspGrid : scanner.bounds().row);
   scanner.reset(&clip);

   int b;
   list<Span> spans;
   while ( scanner.next(b,spans) ) (_accum[b - _grid.min].*initf)(spans);
}

void Machine::loadRegions( struct DataSet *dataset, bool withfg, bool withbg )
{
   string instrument = dataset->attribute("INSTRUME")->asString();

   // initialize _lineSep and _rgsAlpha

   unsigned i = 0;
   const Table *table = dataset->table("SRCLIST");
   const DalVecI index( table->column("INDEX") );
   while ( i < table->rows() && index[i] != _source ) ++i;
   if ( i == table->rows() )
      errstr << "source INDEX=" << _source << " not found in "
	     << table->qualifiedName() << Sas::error("SourceNotFound");

   calServer.setInstrument( Xmm::toInstrument(instrument) );
   struct MiscDataServer *misc = calServer.getAtom(misc);
   _lineSep = misc->get("GRAT_LINE_DENS");
   _rgsAlpha = (misc->get("LINCOORD:LINCOORD%G_RY")
		+ misc->get("INCIDENCE_ANGLE")
		+ (DalVecE(table->column("DELTA_DISP"))[i]/60)
		) * M_PI/180;
   calServer.releaseAtom(misc);

   // initialize the accumulator spans

   char extname[64];

   if ( withbg ) {
      sprintf( extname, "%s_BACKGROUND", instrument.c_str() );
      scan( dataset->table(extname), &SpectralBin::bgSpans, true );
   }

   if ( withfg ) {
      sprintf( extname, "%s_SRC%d_SPATIAL", instrument.c_str(), _source );
      scan( dataset->table(extname), &SpectralBin::fgSpans, true );
   }

   sprintf(extname, "%s_SRC%d_ORDER_%d", instrument.c_str(), _source, _order);
   scan( dataset->table(extname), &SpectralBin::phSpans, false );
}

void Machine::loadExposure( struct Array *array, int edgechannels )
{
   struct ExposureMap emap(array);
   bool valid = false;

   if ( !emap.bounds() )
      errstr << "the exposure map is null";
   else if ( !commensurate(_betaGrid,emap.beta) )
      errstr << "the exposure map dispersion channels are not "
	     << "commensurate with the EVENTS table";
   else if ( !commensurate(_xdspGrid,emap.xdsp) )
      errstr << "the exposure map cross-dispersion channels are not "
	     << "commensurate with the EVENTS table";
   else valid = true;

   if ( !valid ) {
      errstr << ": discarding " << array->qualifiedName()
	     << Sas::warning("InvalidExposureMap");
      return;
   }

   float datamax = array->attribute("DATAMAX")->asReal();
   int ccdnode = array->attribute("CCDNODE")->asInt();
   int ccdid = array->attribute("CCDID")->asInt();
   int nodeId = 2*ccdid + ccdnode - 2;

   const Bounds &bnds = emap.bounds();
   int b;
#if 0
   int offset = bnds.row.min - _xdspGrid.cell( emap.xdsp[bnds.row.min] );
#endif
   for ( int column = bnds.col.min; column <= bnds.col.max; ++column )
      if ( (b = _grid.cell( emap.beta[column] )) >= _grid.min )
	 if ( b <= _grid.max ) {
	    struct SpectralBin &bin = _accum[ b - _grid.min ];
#if 1
	    if ( bin.nodeId == -1 ) bin.nodeId = nodeId;
	    if ( bin.nodeId == nodeId ) {
	       bin.foreground.expose(datamax);
	       bin.background.expose(datamax);
	    }
	    else bin.dubious = true;
#else
	    bin.foreground.expose(emap,column,offset);
	    bin.background.expose(emap,column,offset);
#endif
	 }
	 else break;

   int i, end;
   if ( ccdnode ) {
      i = _grid.cell(emap.beta[bnds.col.min]) - _grid.min;
      end = i + edgechannels;
   }
   else {
      end = 1 + _grid.cell(emap.beta[bnds.col.max]) - _grid.min;
      i = end - edgechannels;
   }
   maximize( i, 0 );
   minimize( end, (int)_accum.size() );
   while ( i < end ) _accum[i++].dubious = true;

   maximize( _maxExposure, (double)datamax );
}

void Machine::loadEvents( const Table *table )
{
   if ( _maxExposure == 0 )
      Sas::error("ZeroExposure","no exposure found in any node");

   const DalVecI beta( table->column("BETA_CHANNEL") );
   const DalVecI xdsp( table->column("XDSP_CHANNEL") );
   const DalVecI calpi( table->column("PI") );

   for ( unsigned i = 0; i < table->rows(); ++i ) {
      int b = (beta[i] - _grid.min) / _rebin;
      if ( !_grid.contains(b+_grid.min) ||
	   !_accum[b].accept(xdsp[i],calpi[i])
	   )
	 errstr << "event " << i+1 << " at (" << beta[i] << "," << xdsp[i]
		<< ") has no associated exposure"
		<< Sas::warning("EventDiscarded");
   }
}

void Machine::out( const string &dsname, const string &hduclas2,
		   const list<const Attribute*> &attrib ) const
{
   struct DataSet *dataset = dataSetServer->open( dsname, Dal::Create );

   list<const Attribute*>::const_iterator itr;
   for ( itr = attrib.begin(); itr != attrib.end(); ++itr )
      dataset->addAttribute(*itr);

   out( dataset, hduclas2, attrib );

   dataSetServer->close( dataset );
}

void Machine::out( struct DataSet *dataset, const string &hduclas2,
		   const list<const Attribute*> &attrib ) const
{
   bool netSpectrum = (hduclas2 == "NET");
   bool bkgSpectrum = (hduclas2 == "BKG");

   struct Table *table = dataset->addTable( "SPECTRUM", _accum.size() );

   list<const Attribute*>::const_iterator itr;
   for ( itr = attrib.begin(); itr != attrib.end(); ++itr )
      table->addAttribute(*itr);

#define ADD(k,r,u,v) *table->addAttribute(k,r,u) = v
   ADD("HDUCLASS","format conforms to OGIP standard","","OGIP");
   ADD("HDUCLAS1","OGIP Memo OGIP/92-007","","SPECTRUM");
   ADD("HDUCLAS2","what combination of source and background","",hduclas2);
   ADD("HDUCLAS3","which data representation","","COUNT");
   ADD("HDUVERS","version of format specification","","1.2.0");
   ADD("POISSERR","whether Poissonian errors may be assumed","",!netSpectrum);
   ADD("CHANTYPE","which channel calibration applies","","PI");
   ADD("DETCHANS","total number of detector channels","",(int)_accum.size());
   ADD("SYS_ERR","no systematic error specified","",0);
   ADD("GROUPING","no channel grouping specified","",0);
   ADD("CORRSCAL","correlation scale factor","",1);
   ADD("CORRFILE","associated correlation file","","none");
   ADD("RESPFILE","associated redistribution matrix","","none");
   ADD("ANCRFILE","associated ancillary response file","","none");
   ADD("BACKFILE","associated background spectrum","","none");
   ADD("SOURCEID","source extracted","SRCLIST:INDEX",_source);
   ADD("RFLORDER","reflection order extracted","",_order);
   ADD("LINE_SEP","grating line spacing","Angstroms",_lineSep);
   ADD("RGSALPHA","grating incidence angle","rad",_rgsAlpha);
   ADD("EXPOSURE","maximum effective integration time","s",_maxExposure);
#undef ADD

   DalVecI channel( table->addColumn("CHANNEL",Dal::Int16,
				     "dispersion channel","") );
   DalVecJ counts( table->addColumn("COUNTS",Dal::Int32,
				    "photon detection count","counts") );
   DalVecE stat_err;
   if ( netSpectrum )
      stat_err.reset( table->addColumn("STAT_ERR",Dal::Real32,
				       "statistical error","counts") );
   DalVecI quality( table->addColumn("QUALITY",Dal::Int16,
				     "quality flag","") );
   DalVecE areascal( table->addColumn("AREASCAL",Dal::Real32,
				      "area scaling factor","") );
   DalVecE backscal;
   if ( !netSpectrum )
      backscal.reset( table->addColumn("BACKSCAL",Dal::Real32,
				       "background scaling factor","") );

   struct Column *chc = channel.column();
   *chc->addAttribute("TCTYP","channel axis type") = "";
   *chc->addAttribute("TCUNI","channel axis units") = "rad";
   *chc->addAttribute("TCRPX","reference channel number") = _grid.min;
   *chc->addAttribute("TCRVL","center of reference channel")= _grid[_grid.min];
   *chc->addAttribute("TCDLT","uniform channel width") = _grid.stride;
   *chc->addAttribute("TLMIN") = _grid.min;
   *chc->addAttribute("TLMAX") = _grid.max;

   struct Statistics fg, bg;
   for ( unsigned i = 0; i < _accum.size(); ++i ) {
      const SpectralBin &bin = _accum[i];
      fg = (bkgSpectrum ? bin.background : bin.foreground).stats(_rebin);

      if ( fg._exposure > 0 ) {
	 float fgT =
	    fg._counts ? fg._mass / fg._counts : fg._exposure / fg._pixels;

	 counts[i] = fg._counts;
	 areascal[i] = fgT / _maxExposure;
	 quality[i] = bin.dubious ? 2 : 0;

	 if ( netSpectrum ) {
	    bg = bin.background.stats(_rebin);
	    if ( bg._exposure > 0 )
	       if ( bg._counts ) {
		  float scale = fg._exposure / bg._exposure;
		  float Q = scale * fgT * bg._counts / bg._mass;
		  counts[i] -= int(.5 + scale * bg._counts);
		  stat_err[i] = sqrt( fg._counts + bg._counts * Q*Q );
	       }
	       else stat_err[i] = fg._counts ? sqrt( fg._counts ) : 1;
	    else {
	       stat_err[i] = fg._counts ? sqrt( fg._counts ) : 1;
	       quality[i] = 2;
	    }
	 }
	 else backscal[i] = fg._exposure / _maxExposure;
      }
      else {
	 counts[i] = 0;
	 areascal[i] = 0;
	 quality[i] = 1;

	 if ( netSpectrum )
	    stat_err[i] = 0;
	 else
	    backscal[i] = 0;
      }

      channel[i] = _grid.min + (int)i;
   }
}

/* --- SpectralBin methods ------------------------------------------------- */

bool SpectralBin::accept( short xdsp, short calpi )
{
   list<Span>::const_iterator itr;
   for ( itr = _banana.begin(); itr != _banana.end(); ++itr )
      if ( calpi <= itr->max )
	 return (calpi < itr->min) ? true
	    : foreground.accept(xdsp) && background.accept(xdsp);
   return true;
}

/* --- Accumulator methods ------------------------------------------------- */

void Accumulator::spans( const list<Span> &spans )
{
   list<Span>::const_iterator itr;
   for ( itr = spans.begin(); itr != spans.end(); ++itr )
      _spans.push_back( AccumSpan(*itr) );
}

void Accumulator::expose( const ExposureMap &emap, int column, int offset )
{
   list<AccumSpan>::iterator itr;
   for ( itr = _spans.begin(); itr != _spans.end(); ++itr )
      itr->expose(emap,column,offset);
}

void Accumulator::expose( float exposure )
{
   list<AccumSpan>::iterator itr;
   for ( itr = _spans.begin(); itr != _spans.end(); ++itr )
      itr->expose(exposure);
}

bool Accumulator::accept( short xdsp )
{
   list<AccumSpan>::iterator itr;
   for ( itr = _spans.begin(); itr != _spans.end(); ++itr )
      if ( xdsp <= itr->max ) return itr->accept(xdsp);
   return true;
}

struct Statistics Accumulator::stats( int rebin ) const
{
   struct Statistics s;
   list<AccumSpan>::const_iterator itr;
   for ( itr = _spans.begin(); itr != _spans.end(); ++itr )
      s += itr->stats(rebin);
   return s;
}

/* --- AccumSpan methods --------------------------------------------------- */

AccumSpan::AccumSpan( const Span &span )
   : Span(span)
{
   _counts.resize( length() );
   _exposure.resize( length() );
}

void AccumSpan::expose( const ExposureMap &emap, int column, int offset )
{
   struct Span span = Span(offset+min,offset+max) & emap.bounds().row;
   if ( span ) {
      long itr = emap.id(column,span.min-1), end = emap.id(column,span.max);
      for ( unsigned i = span.min - (offset+min); itr != end; )
	 _exposure[i++] += emap[ emap.incr(itr,Compass::North) ];
   }
}

void AccumSpan::expose( float exposure )
{
   vector<float>::iterator itr;
   for ( itr = _exposure.begin(); itr != _exposure.end(); ++itr )
      *itr += exposure;
}

bool AccumSpan::accept( short xdsp )
{
   if ( contains(xdsp) ) {
      unsigned i = xdsp - min;
      if ( _exposure[i] <= 0 ) return false;
      ++_counts[i];
   }
   return true;
}

struct Statistics AccumSpan::stats( int rebin ) const
{
   struct Statistics s;
   s._pixels = length();
   for ( unsigned i = 0; i < s._pixels; ++i ) {
      unsigned counts = _counts[i];
      float exposure = _exposure[i] / rebin;
      if ( exposure > 0 ) {
	 s._exposure += exposure;
	 s._mass += counts * exposure;
	 s._counts += counts;
      }
   }
   return s;
}

/* --- Statistics methods -------------------------------------------------- */

Statistics::Statistics()
   : _mass(0), _exposure(0), _counts(0), _pixels(0)
{}

Statistics::Statistics& Statistics::operator +=( const Statistics &s )
{
   _mass += s._mass;
   _exposure += s._exposure;
   _counts += s._counts;
   _pixels += s._pixels;
   return *this;
}
